/*  ACSL specifications for task_kill function from LSM.
    Copyright (C) 2018 Denis Efremov <efremov@ispras.ru>
    Copyright (c) 2018 ISP RAS (http://www.ispras.ru)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "kernel.h"

typedef uint8_t  ilev_t;
typedef uint8_t  lev_t;
typedef uint16_t type_t;
typedef uint64_t cat_t;
typedef struct {
   lev_t  lev;
   ilev_t ilev;
   cat_t  cat;
   type_t type;

   atomic_t ucnt;
} seclabel_t;


// Функция ядра. Атомарная функция уменешения счетчика с проверкой на ноль.
/*@ requires \valid(v);
    assigns v->counter;
    ensures v->counter == \old(v->counter) - 1;
    behavior TRUE:
        assumes v->counter == 1;
        ensures \result == 1;
    behavior FALSE:
        assumes v->counter > 1;
        ensures \result == 0;
    complete behaviors;
    disjoint behaviors;
  */
static inline int atomic_dec_and_test(atomic_t * v);

// Функция ядра. Атомарная функция увеличения счетчика.
/*@ requires \valid(v);
    requires v->counter < INT_MAX;
    assigns v->counter;
    ensures v->counter == \old(v->counter) + 1;
 */
static inline void atomic_inc(atomic_t * v);

// Переменная на уровне спецификаций.
// В ней моделируется вхождение в секцию rcu блокировки.
//@ ghost int ghost_rcu_lock;

// Функция ядра - для работы с rcu блокировками.
// Обозначает вход в критическую секцию rcu.
// На уровне спецификаций отображено только увеличение
// переменной ghost_rcu_lock.
/*@ assigns ghost_rcu_lock;
    ensures ghost_rcu_lock == \old(ghost_rcu_lock) + 1;
*/
static inline void rcu_read_lock(void);


// Проверка, что функция вызывается под rcu блокировкой.
/*@ assigns \nothing;
    behavior IN:
       assumes ghost_rcu_lock != 0;
       ensures \result == 1;
    behavior OUT:
       assumes ghost_rcu_lock == 0;
       ensures \result == 0;
    complete behaviors;
    disjoint behaviors;
*/
static inline int rcu_read_lock_held(void);

// Выход из rcu блокировки.
/*@ assigns ghost_rcu_lock;
    ensures ghost_rcu_lock == \old(ghost_rcu_lock) - 1;
*/
static inline void rcu_read_unlock(void);


/*@ axiomatic CnfCategorySetTheory {
    type CnfCategory = int;
    type CnfCategorySet;

    logic CnfCategorySet EmptyCnfCategorySet; // Функция, возвращающая пустое множество
    logic CnfCategorySet addCnfCategory(CnfCategorySet s, CnfCategory c); // Добавление категории в мнодество
    logic CnfCategorySet intersectCnfCategorySet(CnfCategorySet s1, CnfCategorySet s2); // Перемечение
    logic boolean isinCnfCategorySet(CnfCategorySet s, CnfCategory c); // Проверка вхождения
    logic int sizeCnfCategorySet(CnfCategorySet s); // Мощность множества

    // Примеры лемм

    // Добавление уже существующей категории
    axiom add2Set:
     \forall CnfCategorySet s, CnfCategory c;
      addCnfCategory(addCnfCategory(s,c),c) == addCnfCategory(s,c);

    axiom addCnfCategory_trans:
     \forall CnfCategorySet s, CnfCategory i, CnfCategory j;
      addCnfCategory(addCnfCategory(s,i),j) == addCnfCategory(addCnfCategory(s,j),i);

    // Проверка вхождения в пустое множество
    axiom isinEmptySet:
     \forall CnfCategory c;
      isinCnfCategorySet(EmptyCnfCategorySet,c) == \false;

    axiom isinAddCnfCategory1:
     \forall CnfCategorySet s, CnfCategory c;
      isinCnfCategorySet(addCnfCategory(s,c),c);

    axiom isinAddCnfCategory2:
     \forall CnfCategorySet s, CnfCategory i, CnfCategory j;
      (i != j) ==> (isinCnfCategorySet(addCnfCategory(s,i),j) == isinCnfCategorySet(s,j));

    axiom intersect_com:
     \forall CnfCategorySet s1, CnfCategorySet s2;
      intersectCnfCategorySet(s1,s2) == intersectCnfCategorySet(s2,s1);

    axiom intersect_trans:
     \forall CnfCategorySet s1, CnfCategorySet s2, CnfCategorySet s3;
      intersectCnfCategorySet(s1,intersectCnfCategorySet(s2,s3)) == intersectCnfCategorySet(intersectCnfCategorySet(s1,s2),s3);

    axiom intersect_def0:
     \forall CnfCategorySet s;
     intersectCnfCategorySet(EmptyCnfCategorySet,s) == EmptyCnfCategorySet;

    axiom intersect_def1:
     \forall CnfCategorySet s1, CnfCategorySet s2, CnfCategory c;
       isinCnfCategorySet(s1,c) && isinCnfCategorySet(s2,c) <==> isinCnfCategorySet(intersectCnfCategorySet(s1,s2),c);

    // Размер пустого множества
    axiom sizeCnfCategorySet_def0:
     sizeCnfCategorySet(EmptyCnfCategorySet) == 0;

    axiom sizeCnfCategorySet_def1:
     \forall CnfCategorySet s, CnfCategory e;
       isinCnfCategorySet(s,e) ==> sizeCnfCategorySet(addCnfCategory(s,e)) == sizeCnfCategorySet(s);

    axiom sizeCnfCategorySet_def2:
     \forall CnfCategorySet s, CnfCategory e;
       !isinCnfCategorySet(s,e) ==> sizeCnfCategorySet(addCnfCategory(s,e)) == sizeCnfCategorySet(s)+1;
    }
 */

/*@ axiomatic ModelTypesTheory {
       type Label;
       type CnfLevel  = int;
       type Integrity = LowInt | HighInt;

       // Конструктор модельного типа метки безопасности
       logic Label newLabel(
                        CnfLevel       l,
                        Integrity      il,
                        CnfCategorySet c
                      );
       // Логический функции - "геттеры" для полей модельной метки
       logic CnfLevel        LabelCnfLevel(Label l);
       logic Integrity       LabelIntLevel(Label l);
       logic CnfCategorySet  LabelCnfCategorySet(Label l);

       axiom LabelFieldAccessCnfLevel:
          \forall CnfLevel l,
                  Integrity il,
                  CnfCategorySet c;
              l  == LabelCnfLevel(newLabel(l, il, c));

       axiom LabelFieldAccessIntLevel:
          \forall CnfLevel l,
                  Integrity il,
                  CnfCategorySet c;
              il == LabelIntLevel(newLabel(l, il, c));

       axiom LabelFieldAccessCnfCategorySet:
          \forall CnfLevel l,
                  Integrity il,
                  CnfCategorySet c;
              c  == LabelCnfCategorySet(newLabel(l, il, c));

    }
 */

/*@ axiomatic ModelTypes {
    // Функция преобразования в модельный уровень целостности
    logic Integrity i2m_IntLevel(ilev_t il) = (il == 0) ? LowInt : HighInt;

    // Предикаты сравнения уровней целостности
    predicate IntLevelLt(Integrity i1, Integrity i2) =
       i1 == LowInt && i2 == HighInt;
    predicate IntLevelGt(Integrity i1, Integrity i2) =
       i1 == HighInt && i2 == LowInt;
    predicate IntLevelLte(Integrity i1, Integrity i2) =
       IntLevelLt(i1,i2) || i1 == i2;
    predicate IntLevelGte(Integrity i1, Integrity i2) =
       IntLevelGt(i1,i2) || i1 == i2;

    // Вспомогательная функций
    logic CnfCategorySet i2m_CnfCategorySet64(cat_t c, int n) =
       n == 0 ?
         EmptyCnfCategorySet
       :
         \let cnfset = i2m_CnfCategorySet64((cat_t) (c << (cat_t) 1), (int) (n - (int) 1));
         ((c & (cat_t) 0x1) == 1) ?  addCnfCategory(cnfset, (int) ((int) 64 - n)) : cnfset;
    
    // Функция преобразования в модельное множество категорий
    logic CnfCategorySet i2m_CnfCategorySet(cat_t c) =
       i2m_CnfCategorySet64(c, (int) 64);

    // Функция преобразования в модельную метку безопасности
    logic Label i2m_Label{L}(seclabel_t *l) =
      newLabel(
         l->lev,
         i2m_IntLevel(l->ilev),
         i2m_CnfCategorySet(l->cat)
      );
    }
 */

/*@ axiomatic AccessModes {
      predicate write_bit(int mode) = (mode & MAY_WRITE);
      predicate valid_mode(int mode) = write_bit(mode) || mode == 0;

      lemma MayWriteValid:
         write_bit(MAY_WRITE);
    }
*/

/*@ axiomatic EqualCategories {
      predicate equalCategories = \true; // вырожденный предикат для импорта

      // лемма о том, что если категории не совпадают
      // то и их модельные представления не совпадают
      lemma EqualCategories:
        \forall cat_t scat, cat_t ocat;
        (scat != ocat) ==> (i2m_CnfCategorySet(scat) != i2m_CnfCategorySet(ocat));
    }
*/

/*@ axiomatic AccessIsPermitted {
      predicate AccessIsPermitted{L}(Label s, Label o, int mode);

      axiom AccessIsPermitted_definition:
          \forall int mode, Label s, Label o;
             valid_mode(mode)                                 &&
             write_bit(mode)                                  &&
             LabelCnfLevel(s)       == LabelCnfLevel(o)       &&
             LabelCnfCategorySet(s) == LabelCnfCategorySet(o) &&
             IntLevelGte(LabelIntLevel(s), LabelIntLevel(o))
                  <==>
             AccessIsPermitted(s, o, mode);
    }
*/

/*@ requires equalCategories;   // импорт леммы
    requires valid_seclabel(s);
    requires valid_seclabel(o);
    requires valid_mode(mode);
    assigns \nothing;

    behavior System:
       ensures    \result == 0
               || \result == -EPERM;

    behavior Model:
       ensures \result == 0 <==> AccessIsPermitted(i2m_Label(s),
                                                   i2m_Label(o),
                                                   mode);
*/
int access_is_permitted(const seclabel_t *s, const seclabel_t *o, int mode)
{
   if (mode & MAY_WRITE) {
      if (s->lev  == o->lev
       && s->cat  == o->cat
       && s->ilev >= o->ilev) {
         return 0;
      }
   }

   return -EPERM;
}

/*@ axiomatic SecLabel {

    // Предикат проверки, что в состоянии L2 counter больше на 1
    // чем в состоянии L1.
    predicate seclabel_incremented{L1,L2}(seclabel_t *l) =
      \at(l->ucnt.counter, L1) + 1 == \at(l->ucnt.counter, L2);

    // Предикат проверки, что в состоянии L2 counter меньше на 1
    // чем в состоянии L1.
    predicate seclabel_decremented{L1,L2}(seclabel_t *l) =
      \at(l->ucnt.counter, L1) - 1 == \at(l->ucnt.counter, L2);

    // Предикат проверки возможности захвата ресурса метки безопасности.
    // Содержит требования валидности указателя,
    // возможности увеличение счетчика без переполнения.
    // Требование offset_min нужно для парности с seclabel_put_requires.
    // \offset_min(l) == 0 означает что указатель должен указывать на 
    // начало аллоцированного блока памяти. Это требование для корректности
    // free функций.
    predicate seclabel_get_requires(seclabel_t* l) =
       \valid(l)           &&
       \offset_min(l) == 0 &&
       l->ucnt.counter < INT_MAX;

    predicate seclabel_put_requires(seclabel_t* l) =
       \valid(l)           &&
       \offset_min(l) == 0 &&
       l->ucnt.counter > 0;

    // В случае, если счётчик ссылок опускается до нуля,
    // указатель на метку безопасности становится не валидным.
    // В обратном случае, просто уменьшается счётчик использований.
    predicate seclabel_put_ensures{L1,L2}(seclabel_t* l) =
       ((\at(l->ucnt.counter,L1) == 1) ==> !\at(\valid(\at(l,L1)),L2)) &&
       ((\at(l->ucnt.counter,L1) >  1) ==> seclabel_decremented{L1,L2}(l));

    // Корректность метки безопасности.
    // Среди ранее рассмотренных требований, проверяется что
    // уровень целостности может сожержать только два значения.
    predicate valid_seclabel(seclabel_t *l) =
       \valid(l)                      &&
       \offset_min(l) == 0            &&
       0 < l->ucnt.counter            &&
       (l->ilev == 0 || l->ilev == 1);

    // Проверяет, что в состоянии L1 и L2 counter равны.
    predicate seclabel_counter_unchanged{L1,L2}(seclabel_t *l) =
       \at(l->ucnt.counter, L1) == \at(l->ucnt.counter, L2);

    }
 */

/*@ requires seclabel_get_requires(l); // возможность увеличить счетчик ссылок
    assigns l->ucnt.counter; // меняется только счетчик ссылок
    ensures seclabel_incremented{Pre,Post}(l); // предикат, который проверяет, что счетчик ссылок изменился на 1 в Post по вреынению с Pre состоянием
    ensures \result == l; // возвращается метка безопасности из аргументов
*/
const seclabel_t *seclabel_get(const seclabel_t * l)
{
   seclabel_t *ncl = (seclabel_t *) l;
   atomic_inc(&(ncl->ucnt));

   return l;
}

/*@ requires seclabel_put_requires(l);
    assigns  l->ucnt.counter; // меняется только счетчик ссылок
    ensures  seclabel_put_ensures{Pre,Post}(l); // счетчик ссылок уменьшается на 1

    behavior DEC:
       assumes l->ucnt.counter > 1;
       frees   \nothing;
    behavior FREE:
       assumes l->ucnt.counter == 1;
       frees   l; // освобождается только указатель l

    complete behaviors;
    disjoint behaviors;
*/
void seclabel_put(const seclabel_t * l)
{
   seclabel_t *ncl = (seclabel_t *) l;

   if (!atomic_dec_and_test(&(ncl->ucnt)))
      return;

   kfree(ncl);
}

/*======================================================================*/

/*@ axiomatic TaskLabel {
      // Предикат корректности структуры cred.
      // Требует валидности указателя, накладывает ограничение
      // на тип поля security. Требует, чтобы поле security было
      // всегда инициализировано (это требование для упрощения примера).
      // Требует, чтобы счетчик ссылок был больше нуля.
      predicate valid_cred(struct cred *cred) =
        \valid(cred)                                     &&
        (\typeof(cred->security) <: \type(seclabel_t *)) &&
        valid_seclabel((seclabel_t *)cred->security)     &&
        0 < cred->usage.counter;

      // Предикат корректности структуры task_struct требует
      // валидности указателя, корректности cred и real_cred
      // структур через предикат valid_cred, совпадения
      // указателей на метку безопасности у cred и valid_cred.
      predicate valid_task_struct(struct task_struct* tsk) =
        \valid(tsk)                &&
        valid_cred(tsk->cred)      &&
        valid_cred(tsk->real_cred) &&
        tsk->cred->security == tsk->real_cred->security;

      // Логическая функция TaskLabel всегда берет метку
      // безопаности из субъектных прав cred, так мы считаем
      // что security поле у cred и real_cred всегда совпадает.
      // Это ограничение отражено в предикате корретности метки
      // безопасности valid_seclabel.
      logic seclabel_t *TaskLabel{L}(struct task_struct *t) =
        (seclabel_t *)(t->cred->security);
    }
*/

/*@ requires valid_task_struct(task);    // требование корректности структур данных
    requires valid_task_struct(current);
    requires task != current && task->cred != cred; // в рамках примера мы считаем, что процесс не может послать сигнал себе сам
    requires valid_seclabel(TaskLabel(task)); // требование корректности метки
    requires seclabel_get_requires(TaskLabel(task)); // требование возможности захвата ресурса
    requires valid_cred(cred) && // если аргумент cred задан
             (\let sbj_sec = (seclabel_t *)cred->security; // метка безопасности берется из cred
              valid_seclabel(sbj_sec)                      &&
              sbj_sec != TaskLabel(task)                   && // метки всегда различны
              seclabel_get_requires(sbj_sec))
          ||
             cred == \null && // если аргумент cred равен NULL
             (\let sbj_sec = TaskLabel(current);  // метка безопасности берется у процесса current
              valid_seclabel(sbj_sec)            &&
              sbj_sec != TaskLabel(task)         && // метки всегда различны
              seclabel_get_requires(sbj_sec));

    // описывается множество, за пределами которого не меняется память
    assigns TaskLabel(task)->ucnt,
            ((seclabel_t *)cred->security)->ucnt,
            TaskLabel(current)->ucnt,
            ghost_rcu_lock;

    behavior System:
       ensures seclabel_counter_unchanged{Pre,Post}(TaskLabel(task)); // счетчик ссылок у ресурса не изменился
       ensures cred == \null ==>
               seclabel_counter_unchanged{Pre,Post}(TaskLabel(current));
       ensures valid_cred(cred) ==>
               seclabel_counter_unchanged{Pre,Post}((seclabel_t *)cred->security);
       ensures    \result == 0
               || \result == -EPERM;

    behavior ModelCred:
       assumes cred != \null;
       ensures \result == 0 <==> AccessIsPermitted(
                                    i2m_Label((seclabel_t *)cred->security),
                                    i2m_Label(TaskLabel(task)),
                                    MAY_WRITE);

    // cred равен NULL, проверка для метки безопасности  провесса current
    behavior ModelCurrent:
       assumes cred == \null;
       ensures \result == 0 <==> AccessIsPermitted(
                                    i2m_Label(TaskLabel(current)),
                                    i2m_Label(TaskLabel(task)),
                                    MAY_WRITE);

    // исходя из предусловий, assumes в ModelCred и ModelCurrent не могут быть верны одновременно
    disjoint behaviors ModelCred, ModelCurrent;
    // исходя из предусловий, выполнено хотя бы одно из поведений ModelCred, ModelCurrent
    complete behaviors ModelCred, ModelCurrent;
 */
int mylsm_kill_permission(struct task_struct *task,
                          struct siginfo *info,
                          int sig, const struct cred *cred)
{
   int rc;
   const seclabel_t *slbl_task, *slbl_curr;

   /* sig == 0 в примере не рассматривается как MAY_READ доступ */
   /* Считается, что метка безопасности всегда проинициализирована */

   /* Subjective credentials: cred */
   if (cred) {
      //@ assert ((seclabel_t*)cred->security)->ucnt != ((seclabel_t *)task->cred->security)->ucnt;
      slbl_curr = seclabel_get(cred->security);
   } else {
      //@ assert ((seclabel_t *)current->cred->security)->ucnt != ((seclabel_t *)task->cred->security)->ucnt;
      slbl_curr = seclabel_get(current_security()); /* cred */
   }

   /* Objective credentials: real_cred */
   slbl_task = seclabel_get(task_security(task));

   rc = access_is_permitted(slbl_curr, slbl_task, MAY_WRITE);

   seclabel_put(slbl_task);
   seclabel_put(slbl_curr);

   return rc;
}

/* vim: set et sts=0 sw=3 ts=3 : */
